/*****************************************************************************/
/*
                               HyperShelf.c

A CGI-compliant script.

Emulates the "Bookshelf Navigation Utility" (BNU) of 1997ff on-line
documentation CD releases, and the Bookreader shelf navigation capability of
earlier times. It has some additional capabilities not found in either, the
ability to provide access to PostScript and plain-text documents.

Bookreader supports Bookreader format documents.

BNU (currently) supports Bookreader and HTML documents.

Hypershelf externsions: PostScript, plain-text and Adobe PDF documents.

Note that the Hypershelf extensions are preceded by two comment characters,
either "#" (as in the example) or "!".  The prevents the BNU from barfing on
them.  If a title is the same as the previous entry (or is missing altogether)
the document title becomes a description of the document type.  For examples of
how they behave do a 

  http:///hypershelf/ht_root/exercise/library.odl

  http:///hypershelf/ht_root/exercise/library.decw$bookshelf

Here is an example BNU shelf:

-------------------------------------------------------------------------------
title	VMS Hypertext Services Library (BNU version!)
book	/ht_root/doc/htd/htd.decw$book	Technical Overview
html	/ht_root/doc/htd/htd_0000.html	Technical Overview
##post	/ht_root/doc/htd/htd.ps		Technical Overview
##plain	/ht_root/doc/htd/htd.txt	Technical Overview
##pdf	/ht_root/doc/htd/htd.pdf	Technical Overview
book	/ht_root/doc/env/env.decw$book	Hypertext Environment
html	/ht_root/doc/env/env_0000.html	Hypertext Environment
##post	/ht_root/doc/env/env.ps	 	Hypertext Environment
##plain	/ht_root/doc/env/env.txt	Hypertext Environment
book	/ht_root/doc/sdm2htm/sdm2htm.decw$book	SDML to HTML Converter
html	/ht_root/doc/sdm2htm/sdm2htm_0000.html	SDML to HTML Converter
##post	/ht_root/doc/sdm2htm/sdm2htm.ps	SDML to HTML Converter
book	/ht_root/doc/menu_primer/menu_primer.decw$book	Menu Primer
html	/ht_root/doc/menu_primer/menu_primer_0000.html	Menu Primer
-------------------------------------------------------------------------------

Here is an example Bookreader shelf:

-------------------------------------------------------------------------------
title\\VMS Hypertext Services Library (Bookreader version!)
book\ht_root:[doc.htd]htd\Technical Overview
##html\ht_root:[doc.htd]htd_0000.html\
##ps\ht_root:[doc.htd]htd.ps\
##plain\ht_root:[doc.htd]htd.txt\
##pdf\ht_root:[doc.htd]htd.pdf\
book\ht_root:[doc.env]env\Hypertext Environment
##html\ht_root:[doc.env]env_0000.html\
##post\ht_root:[doc.env]env.ps\
##plain\ht_root:[doc.env]env.txt\
book\ht_root:[doc.sdm2htm]sdm2htm\SDML to HTML Converter
##html\ht_root:[doc.sdm2htm]sdm2htm_0000.html\
##ps\ht_root:[doc.sdm2htm]sdm2htm.ps\
book\ht_root:[doc.menu_primer]menu_primer\Menu Primer
##html\ht_root:[doc.menu_primer]menu_primer_0000.html\Menu Primer
-------------------------------------------------------------------------------


FORM FIELDS
-----------
help=           if non-null causes a redirection to the help HTML document
referer=        if non-null provides referer URL for closing library
                (must have any URL-escape percentages URL-escaped!)
shelf=          the name of the shelf to be opened
title=          the title (description) of the shelf being opened


QUALIFIERS
----------
/BODY=          <BODY> tag attributes (e.g. background colour, text colour)
/BOOKREADER     Bookreader is the default shelf processor
/BNU            Bookshelf Navigation Utility is the default shelf processor
/DBUG           turns on all "if (Debug)" statements
/HELP=          URL of help HTML file (optional)
/HYPERREADER=   name of 'HyperReader' script (optional)
/ICON=          URL location of the icons
/LIBRARY=       specifies the default library (shelf) file
/PRINT=         name of 'Print' script (optional)


LOGICAL NAMES
-------------
HTTPD$GMT       timezone offset from Greenwich Mean Time (e.g. "09:30")


BUILD DETAILS
-------------
See BUILD_HYPERSHELF.COM procedure.


VERSION HISTORY (update SoftwareID as well!)
---------------
01-AUG-97  MGD  v3.2.1, 'HttpHasBeenOutput' not initialize for CGIplus
20-JUL-97  MGD  v3.2.0, added /BODY= qualifier, added Adobe PDF document type
07-JUL-97  MGD  v3.1.0, CGIplus capable
30-JUN-97  MGD  v3.0.0, major changes to support the BNU format shelves,
                        "Pragma: no-cache" now overrides "If-Modified-Since:"
16-MAY-96  MGD  v2.2.3, changes all .XBMs to transparent .GIFs
23-FEB-96  MGD  v2.2.2, bugfix, after modifying the HTTPD$GMT format for the
                        HTTPd server I had forgotten about some scripts
22-NOV-95  MGD  v2.2.1, discovered the TITLE keyword in libraries
12-OCT-95  MGD  v2.2.0, added 'Referer:', 'Last-Modified:'/'If-Modified-Since:'
24-MAY-95  MGD  v2.1.0, minor changes for AXP compatibility
15-APR-95  MGD  v2.0.0, complete rewrite
10-JUN-94  MGD  v1.0.0, initial development
*/
/*****************************************************************************/

#ifdef __ALPHA
   char SoftwareID [] = "HYPERSHELF AXP-3.2.1";
#else
   char SoftwareID [] = "HYPERSHELF VAX-3.2.1";
#endif

/* standard C header files */
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <unixio.h>
#include <time.h>

/* VMS-related header files */
#include <descrip.h>
#include <libdef.h>
#include <lnmdef.h>
#include <rms.h>
#include <ssdef.h>
#include <stsdef.h>

#define boolean int
#define true 1
#define false 0

#ifdef __ALPHA
#   pragma nomember_alignment
#endif

#define VMSok(x) ((x) & STS$M_SUCCESS)
#define VMSnok(x) (!((x) & STS$M_SUCCESS))
 
#define DefaultBnuLibrary "DECW$SYSTEM_DEFAULTS:LIBRARY.ODL"
#define DefaultDecwBookshelf "DECW$BOOKSHELF"
#define DecwBook "DECW$BOOK:"

char *ShelfFileTypes [] = { ".ODL", ".DECW$BOOKSHELF", "" };

char SgmlDocType [] = "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2//EN\">";

char  Http302Header [] =
"HTTP/1.0 302 Redirection.\r\n\
Location: http://%s%s\r\n\
\r\n";

char  Http404Header [] =
"HTTP/1.0 404 Error\r\n\
Content-Type: text/html\r\n\
\r\n";

char  Utility [] = "HYPERSHELF";

boolean  Debug,
         DoBookreader,
         DoBnu,
         ErrorReported,
         HttpHasBeenOutput,
         IsCgiPlus,
         TimeAheadOfGmt;

unsigned long  IfModifiedSinceBinaryTime [2],
               TimeGmtDeltaBinary [2];

int  ShelfSpecLength;

char  *CgiPlusEofPtr,
      *CgiTypePtr;

char  CgiHttpIfModifiedSince [64],
      CgiHttpPragma [256],
      CgiHttpReferer [256],
      CgiPathInfo [256],
      CgiPathTranslated [256],
      CgiServerName [256],
      CgiScriptName [256],
      CgiServerName [256],
      FormTitle [256],
      GmDateTime [32],
      HelpUrl [256],
      HtmlBodyTag [256],
      HyperReaderScriptName [256],
      HyperShelfIcon [256], 
      IconLocation [256],
      LibraryFileName [256],
      LastModifiedGmDateTime [32],
      OriginalReferer [256],
      PrintScriptName [256], 
      HyperReaderReferer [512],
      TimeGmtString [32],
      TimeGmtVmsString [32],
      UriReferer [512],
      ShelfSpec [256];

/* required function prototypes */
char* CgiVar (char*);
char* MapUrl (char*, char*, char*, char*);

/*****************************************************************************/
/*
*/

main (int argc, char *argv[])

{
   register int  acnt;
   register char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (getenv ("HYPERSHELF$DBUG") != NULL) Debug = true;

   if (Debug)
      system ("show sym *");
   else
   {
      /* reopen output stream so that the '\r' and '\n' are not filtered */
#ifdef __DECC
      if ((stdout = freopen ("SYS$OUTPUT", "w", stdout, "ctx=bin")) == NULL)
         exit (vaxc$errno);
#else
      if ((stdout = freopen ("SYS$OUTPUT", "w", stdout, "rfm=udf")) == NULL)
         exit (vaxc$errno);
#endif
   }

   /***********************************/
   /* get the command line parameters */
   /***********************************/

   /* using this mechanism parameters must be space-separated! */
   for (acnt = 1; acnt < argc; acnt++)
   {
      if (Debug) fprintf (stdout, "argv[%d] |%s|\n", acnt, argv[acnt]);
      if (strsame (argv[acnt], "/DBUG", -1))
      {
         Debug = true;
         continue;
      }
      if (strsame (argv[acnt], "/BODY=", 4))
      {
         for (cptr = argv[acnt]+5; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         sptr = HtmlBodyTag;
         if (*cptr) *sptr++ = ' ';
         while (*cptr) *sptr++ = *cptr++;
         *sptr = '\0';
         continue;
      }
      if (strsame (argv[acnt], "/BOOKREADER", 4))
      {
         DoBookreader = true;
         DoBnu = false;
         continue;
      }
      if (strsame (argv[acnt], "/BNU", -1))
      {
         DoBookreader = false;
         DoBnu = true;
         continue;
      }
      if (strsame (argv[acnt], "/HELP=", 4))
      {
         for (cptr = argv[acnt]+5; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         sptr = HelpUrl;
         while (*cptr) *sptr++ = *cptr++;
         *sptr = '\0';
         continue;
      }
      if (strsame (argv[acnt], "/HYPERREADER=", 4))
      {
         for (cptr = argv[acnt]+5; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         sptr = HyperReaderScriptName;
         if (*cptr != '/') *sptr++ = '/';
         while (*cptr) *sptr++ = *cptr++;
         *sptr = '\0';
         continue;
      }
      if (strsame (argv[acnt], "/ICON=", 4))
      {
         for (cptr = argv[acnt]+5; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         sptr = IconLocation;
         while (*cptr) *sptr++ = *cptr++;
         if (sptr > IconLocation)
            if (sptr[-1] == '/')
               sptr--;
         *sptr = '\0';
         sprintf (HyperShelfIcon,
            "<IMG ALIGN=top SRC=\"%s/HyperShelf.gif\" ALT=\"HyperShelf - \"> ",
            IconLocation);
         continue;
      }
      if (strsame (argv[acnt], "/LIBRARY=", 4))
      {
         for (cptr = argv[acnt]+5; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         sptr = LibraryFileName;
         while (*cptr) *sptr++ = *cptr++;
         *sptr = '\0';
         continue;
      }
      if (strsame (argv[acnt], "/PRINT=", 4))
      {
         for (cptr = argv[acnt]+5; *cptr && *cptr != '='; cptr++);
         if (*cptr) cptr++;
         sptr = PrintScriptName;
         if (*cptr != '/') *sptr++ = '/';
         while (*cptr) *sptr++ = *cptr++;
         *sptr = '\0';
         continue;
      }
      fprintf (stdout, "%%%s-E-IVQUAL, unrecognized qualifier\n \\%s\\\n",
               Utility, argv[acnt]+1);
      exit (STS$K_ERROR | STS$M_INHIB_MSG);
   }

   /* if not forced to be BNU and old bookreader logicals define then old */
   if (!DoBnu && (getenv ("DECW$BOOK") != NULL ||
                  getenv ("DECW$BOOKSHELF") != NULL))
      DoBookreader = true;
   else
      DoBnu = true;

   /***********/
   /* process */
   /***********/

   if (IsCgiPlus = ((CgiPlusEofPtr = getenv("CGIPLUSEOF")) != NULL))
      CgiTypePtr = "plus";
   else
      CgiTypePtr = "standard";

   do {

      ProcessRequest ();

      if (IsCgiPlus)
      {
         /* flush, write end-of-output line, flush */
         fflush (stdout);
         if (IsCgiPlus) fprintf (stdout, "%s\n", CgiPlusEofPtr);
         fflush (stdout);
      }

   } while (IsCgiPlus);

   exit (SS$_NORMAL);
}

/*****************************************************************************/
/*
*/ 
 
ProcessRequest ()

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ProcessRequest()\n");

   HttpHasBeenOutput = ErrorReported = false;

   /* if CGIplus, block waiting for next request */
   CgiVar ("");

   /* script only handles GET method */
   if (strcmp (CgiVar ("WWW_REQUEST_METHOD"), "GET")) return;

   if (*CgiVar ("WWW_FORM_HELP"))
   {
      strcpy (CgiServerName, CgiVar ("WWW_SERVER_NAME"));
      fprintf (stdout, Http302Header, CgiServerName, HelpUrl);
      return;
   }

   strcpy (CgiPathInfo, CgiVar ("WWW_PATH_INFO"));
   strcpy (CgiPathTranslated, CgiVar ("WWW_PATH_TRANSLATED"));
   strcpy (CgiScriptName, CgiVar ("WWW_SCRIPT_NAME"));
   strcpy (CgiServerName, CgiVar ("WWW_SERVER_NAME"));
   strcpy (CgiHttpIfModifiedSince, CgiVar ("WWW_HTTP_IF_MODIFIED_SINCE"));
   strcpy (CgiHttpPragma, CgiVar ("WWW_HTTP_PRAGMA"));

   strcpy (ShelfSpec, CgiVar ("WWW_FORM_SHELF"));
   strcpy (FormTitle, CgiVar ("WWW_FORM_TITLE"));

   if (VMSnok (status = TimeSetGmt ()))
   {
      if (status != SS$_NOLOGNAM)
      {
         ErrorVmsStatus (status, CgiPathInfo, CgiPathTranslated,
                         __FILE__, __LINE__);
         return;
      }
   }
   if (CgiHttpIfModifiedSince[0])
   {
      if (VMSnok (HttpGmTime (CgiHttpIfModifiedSince, 
                              &IfModifiedSinceBinaryTime)))
      {
         if (Debug) fprintf (stdout, "If-Modified-Since: NBG!\n");
         IfModifiedSinceBinaryTime[0] = IfModifiedSinceBinaryTime[0] = 0;
         CgiHttpIfModifiedSince[0] = '\0';
      }
   }

   if (CgiPathInfo[0] == '/' && !CgiPathInfo[1])
      CgiPathInfo[0] = CgiPathTranslated[0] = '\0';

   if (!ShelfSpec[0]) strcpy (ShelfSpec, CgiPathTranslated);
   if (!ShelfSpec[0]) strcpy (ShelfSpec, CgiPathInfo);
   if (!ShelfSpec[0]) strcpy (ShelfSpec, LibraryFileName);
   if (!ShelfSpec[0] && DoBnu) strcpy (ShelfSpec, DefaultBnuLibrary);
   if (!ShelfSpec[0]) strcpy (ShelfSpec, DefaultDecwBookshelf);
   ShelfSpecLength = strlen(ShelfSpec);

   if (!FormTitle[0]) strcpy (FormTitle, "Library");

   strcpy (OriginalReferer, CgiVar ("WWW_FORM_REFERER"));

   if (!OriginalReferer[0])
      strcpy (OriginalReferer, CgiVar ("WWW_HTTP_REFERER"));
   if (OriginalReferer[0])
   {
      /* re-escape the URL-escaped percentages */
      CopyIntoUri (UriReferer, OriginalReferer, -1);
   }

   HttpGmTimeString (GmDateTime, 0);

   ProcessShelf ();
}

/*****************************************************************************/
/*
Open the supplied shelf file name (in 'ShelfSpec').  Read each line in the 
file parsing the entry type (shelf or book), the file specification, and entry 
description (title of shelf or book).  Close the file.  If a file 
specification (shelf or book) does not explictly provide a directory 
component then use the directory of the currently open shelf file.
*/ 
 
ProcessShelf ()

{
   register char  *cptr, *sptr;

   boolean  LibraryTitleOutput,
            IsDecwShelf,
            IsOdlShelf,
            IsSameAsPrevTitle;
   int  status,
        idx,
        Length,
        LineCount,
        LineOutputCount;
   char  ExpandedFileName [256],
         FileName [256],
         HtmlTitle [256],
         Line [256],
         LinkFileName [256],
         ParseDefaults [256],
         PrevTitle [256],
         Scratch [512],
         ShelfDirectory [256],
         Title [256],
         Type [256],
         UnixDateTime [32],
         UriTitle [256],
         UriFormTitle [256];
   char*  PossibleOdlShelfTitlePtr;
   unsigned long  UnixTime;
   struct tm  *UnixTmPtr;
   struct FAB  ShelfFab;
   struct NAM  ShelfNam;
   struct RAB  ShelfRab;
   struct XABDAT  ShelfXabDat;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ProcessShelf() |%s|\n", ShelfSpec);

   time (&UnixTime);
   UnixTmPtr = localtime (&UnixTime);
   if (!strftime (UnixDateTime, sizeof(UnixDateTime),
                  "%a, %d %b %Y %T", UnixTmPtr))
      strcpy (UnixDateTime, "[error]");
   if (Debug) fprintf (stdout, "UnixDateTime |%s|\n", UnixDateTime);

   LibraryTitleOutput = false;
   HtmlTitle[0] = UriFormTitle[0] = '\0';

   sptr = HyperReaderReferer;
   sptr += CopyIntoUri (sptr, CgiScriptName, -1);
   *sptr++ = '/';
   sptr += CopyIntoUri (sptr, CgiPathInfo, -1);
   sptr += CopyIntoUri (sptr, "?title=", -1);
   /*
      This escapes the escape percentages leaving the URI still escaped
      when the script (HyperReader) receives it at the other end :^)
   */
   CopyIntoUri (Scratch, FormTitle, -1);
   sptr += CopyIntoUri (sptr, Scratch, -1);
   sptr += CopyIntoUri (sptr, "&referer=", -1);
   sptr += CopyIntoUri (sptr, UriReferer, -1);
   *sptr = '\0';
   if (Debug) fprintf (stdout, "HyperReaderReferer |%s|\n", HyperReaderReferer);

   IsDecwShelf = IsOdlShelf = false;
   for (cptr = ShelfSpec + ShelfSpecLength;
        cptr > ShelfSpec && *cptr != '.';
        *cptr--);
   if (*cptr == '.')
   {
      if (strsame (cptr, ".ODL", -1) ||
          strsame (cptr, ".ODL;", 5))
         IsOdlShelf = true;

      if (strsame (cptr, ".DECW$BOOKSHELF", -1) ||
          strsame (cptr, ".DECW$BOOKSHELF;", 16))
         IsDecwShelf = true;
   }

   /**************/
   /* open shelf */
   /**************/

   for (idx = 0; ShelfFileTypes[idx][0]; idx++)
   {
      if (IsOdlShelf || CgiPathTranslated[0])
         Length = sprintf (ParseDefaults, "%s", ShelfFileTypes[idx]);
      else
         Length = sprintf (ParseDefaults, "%s%s",
                           DecwBook, ShelfFileTypes[idx]);

      ShelfFab = cc$rms_fab;
      ShelfFab.fab$l_dna = ParseDefaults;  
      ShelfFab.fab$b_dns = Length;
      ShelfFab.fab$b_fac = FAB$M_GET;
      ShelfFab.fab$l_fna = ShelfSpec;  
      ShelfFab.fab$b_fns = ShelfSpecLength;
      ShelfFab.fab$l_nam = &ShelfNam;  
      ShelfFab.fab$b_shr = FAB$M_SHRGET;
      ShelfFab.fab$l_xab = &ShelfXabDat;

      ShelfNam = cc$rms_nam;
      ShelfNam.nam$l_esa = ExpandedFileName;
      ShelfNam.nam$b_ess = sizeof(ExpandedFileName)-1;

      ShelfXabDat = cc$rms_xabdat;

      status = sys$open (&ShelfFab, 0, 0);

      /* try with the next extension (if any) */
      if (status == RMS$_FNF) continue;
   }

   if (VMSnok (status))
   {
      if (Debug) fprintf (stdout, "sys$open() %%X%08.08X\n", status);
      if (HtmlTitle[0])
         cptr = HtmlTitle;
      else
      if (CgiPathInfo[0])
         cptr = CgiPathInfo;
      else
         cptr = ShelfSpec;
      ErrorVmsStatus (status, cptr, ShelfSpec, __FILE__, __LINE__);
      return (SS$_NORMAL);
   }

   if (CgiHttpIfModifiedSince[0])
   {
      if (VMSnok (ModifiedSince (&IfModifiedSinceBinaryTime,
                                 &ShelfXabDat.xab$q_rdt)))
      {
         /* book has not been modified since the date/time, don't send */
         sys$close (&ShelfFab, 0, 0);
         return (SS$_NORMAL);
      }
   }

   HttpGmTimeString (LastModifiedGmDateTime, &ShelfXabDat.xab$q_rdt);

   ShelfRab = cc$rms_rab;
   ShelfRab.rab$l_fab = &ShelfFab;
   /* 2 buffers and read ahead performance option */
   ShelfRab.rab$b_mbf = 2;
   ShelfRab.rab$l_rop = RAB$M_RAH;
   ShelfRab.rab$l_ubf = Line;
   ShelfRab.rab$w_usz = sizeof(Line)-1;

   if (VMSnok (status = sys$connect (&ShelfRab, 0, 0)))
   {
      if (Debug) fprintf (stdout, "sys$connect() %%X%08.08X\n", status);
      sys$close (&ShelfFab, 0, 0);
      if (HtmlTitle[0])
         cptr = HtmlTitle;
      else
         cptr = CgiPathInfo;
      ErrorVmsStatus (status, cptr, ShelfSpec, __FILE__, __LINE__);
      return (SS$_NORMAL);
   }

   *ShelfNam.nam$l_ver = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", ShelfNam.nam$l_dev);

   if (strsame (ShelfNam.nam$l_type, ".ODL", -1))
   {
      IsDecwShelf = false;
      IsOdlShelf = true;
   }
   else
   {
      IsDecwShelf = true;
      IsOdlShelf = false;
   }

   /***************/
   /* HTTP header */
   /***************/

   if (!HyperShelfIcon[0]) strcpy (HyperShelfIcon, "<I>HyperShelf</I> ");

   fprintf (stdout,
"HTTP/1.0 200 Document follows.\r\n\
Date: %s\r\n\
Last-Modified: %s\r\n\
Content-Type: text/html\r\n\
\r\n",
   GmDateTime, LastModifiedGmDateTime);
   fflush (stdout);
   HttpHasBeenOutput = true;

   /****************/
   /* file records */
   /****************/

   LineCount = LineOutputCount = 0;
   PrevTitle[0] = '\0';

   while (VMSok (status = sys$get (&ShelfRab, 0, 0)))
   {
      LineCount++;
      Line[ShelfRab.rab$w_rsz] = '\0';
      if (Debug) fprintf (stdout, "Line |%s|\n", Line);
      if (!Line[0]) continue;

      /********************/
      /* parse components */
      /********************/

      cptr = Line;
      if ((cptr[0] == '#' && cptr[1] == '#' && isalpha(cptr[2])) ||
          (cptr[0] == '!' && cptr[1] == '!' && isalpha(cptr[2])))
         cptr += 2;

      /* ignore unknown line types */
      if (!isalpha(*cptr)) continue;

      Type[0] = FileName[0] = Title[0] = '\0';

      sptr = Type;
      while (*cptr && isspace(*cptr)) cptr++;
      if (IsOdlShelf)
      {
         while (*cptr && !isspace(*cptr)) *sptr++ = toupper(*cptr++);
         *sptr = '\0';
      }
      else
      {
         while (*cptr && *cptr != '\\') *sptr++ = toupper(*cptr++);
         *sptr = '\0';
         if (*cptr == '\\') cptr++;
      }

      while (*cptr && isspace(*cptr)) cptr++;
      PossibleOdlShelfTitlePtr = cptr;
      sptr = FileName;
      if (IsOdlShelf)
      {
         while (*cptr && !isspace(*cptr))
         {
            *sptr++ = toupper(*cptr++);
            if (*cptr == '%')
            {
               *sptr++ = '2';
               *sptr++ = '5';
            }
         }
         *sptr = '\0';
      }
      else
      {
         while (*cptr && *cptr != '\\') *sptr++ = toupper(*cptr++);
         *sptr = '\0';
         if (*cptr == '\\') cptr++;
      }

      while (*cptr && isspace(*cptr)) cptr++;
      sptr = Title;
      while (*cptr) *sptr++ = *cptr++;
      *sptr = '\0';

      /***************/
      /* shelf title */
      /***************/

      if (Type[0] == 'T')
      {
         /* title of this shelf */
         if (IsOdlShelf)
         {
            sptr = Title;
            for (cptr = PossibleOdlShelfTitlePtr; *cptr; *sptr++ = *cptr++);
            *sptr = '\0';
            cptr = "ODL";
         }
         else
            cptr = "DECW$BOOKSHELF";
         if (!Title[0])
            sprintf (Title, "[Unable to parse %s shelf title (line %d)]",
                     cptr, LineCount);
         if (LibraryTitleOutput) continue;
         if (!HtmlTitle[0]) CopyIntoHtml (HtmlTitle, Title, -1);
         if (!UriFormTitle[0]) CopyIntoUri (UriFormTitle, Title, -1);
      }

      if (!LibraryTitleOutput)
      {
         if (!HtmlTitle[0]) CopyIntoHtml (HtmlTitle, FormTitle, -1);
         if (!UriFormTitle[0]) CopyIntoUri (UriFormTitle, FormTitle, -1);

         fprintf (stdout,
"%s\n\
<HTML>\n\
<HEAD>\n\
<META NAME=\"generator\" CONTENT=\"%s\">\n\
<META NAME=\"source\" CONTENT=\"%s\">\n\
<META NAME=\"date\" CONTENT=\"%s\">\n\
<META NAME=\"CGI\" CONTENT=\"%s\">\n\
<TITLE>HyperShelf - %s</TITLE>\n\
</HEAD>\n\
<BODY%s>\n\
<TABLE CELLPADDING=0 BORDER=0>\n\
<TR>\n\
<TD VALIGN=top>%s</TD>\n\
<TD><H1>%s</H1></TD>\n\
</TR>\n\
</TABLE>\n",
         SgmlDocType, SoftwareID, CgiServerName, UnixDateTime, CgiTypePtr,
         HtmlTitle, HtmlBodyTag, HyperShelfIcon, HtmlTitle);
         fflush (stdout);

         if (OriginalReferer[0])
            fprintf (stdout,
"<A HREF=\"%s\"><IMG SRC=\"%s/close.gif\" ALT=\"[Close]\"></A>\n",
                     OriginalReferer, IconLocation);

         if (HelpUrl[0])
            fprintf (stdout,
"<A HREF=\"%s\"><IMG SRC=\"%s/help.gif\" ALT=\"[Help]\"></A>\n",
                     HelpUrl, IconLocation);

         fprintf (stdout,
"<BR><IMG SRC=\"%s/hr.gif\" ALT=\"%s\">\n<P>",
            IconLocation, "+--------------------------------------+");
         LibraryTitleOutput = true;

         fprintf (stdout, "<TABLE CELLPADDING=2 BORDER=0>\n");
      }

      if (Type[0] == 'T') continue;

      /*********************/
      /* shelf or document */
      /*********************/

      if (!FileName[0]) strcpy (FileName, "ERROR-IN-SHELF-ENTRY");

      CopyIntoHtml (HtmlTitle, Title, -1);
      CopyIntoUri (UriTitle, Title, -1);

      if (strsame (Title, PrevTitle, -1) || strsame (Title, "ditto", 5))
         IsSameAsPrevTitle = true;
      else
         IsSameAsPrevTitle = false;

      if (IsOdlShelf)
      {
         /*************/
         /* ODL shelf */
         /*************/

         if (FileName[0] == '%')
         {
            /* full path was specified */
            sptr = LinkFileName;
            for (cptr = "/disk$"; *cptr; *sptr++ = *cptr++);
            for (cptr = FileName+1; *cptr; *sptr++ = tolower(*cptr++));
            *sptr = '\0';
         }
         else
         {
            sptr = LinkFileName;
            for (cptr = FileName; *cptr; *sptr++ = tolower(*cptr++));
            *sptr = '\0';
         }
         if (Debug) fprintf (stdout, "LinkFileName |%s|\n", LinkFileName);
      }
      else
      {
         /************************/
         /* DECW$BOOKSHELF shelf */
         /************************/

         sptr = LinkFileName;
         for (cptr = FileName; *cptr; cptr++) *sptr++ = tolower(*cptr);
         *sptr = '\0';
      }

      /*******************/
      /* output the HTML */
      /*******************/

      LineCount++;

      fprintf (stdout, "<TR><TD><NOBR>\n");

      if (Type[0] == 'S')
      {
         /*********/
         /* shelf */
         /*********/

         if (IsOdlShelf || LinkFileName[0] == '/')
         {
            if (IsOdlShelf)
            {
               for (cptr = LinkFileName; *cptr; cptr++);
               while (cptr > LinkFileName && *cptr != '.') cptr--;
               if (*cptr != '.') strcat (LinkFileName, ".odl");
            }
            else
            {
               for (cptr = LinkFileName; *cptr; cptr++);
               while (cptr > LinkFileName && *cptr != '.') cptr--;
               if (*cptr != '.') strcat (LinkFileName, ".decw$bookshelf");
            }

            fprintf (stdout,
"<A HREF=\"%s?title=%s&referer=%s\">\
<IMG SRC=\"%s/shelf.gif\" ALIGN=top ALT=\"_|\\_\"></A>\n",
            LinkFileName, UriTitle, UriReferer, IconLocation);
         }
         else
         {
            fprintf (stdout,
"<A HREF=\"%s?shelf=%s&title=%s&referer=%s\">\
<IMG SRC=\"%s/shelf.gif\" ALIGN=top ALT=\"_|\\_\"></A>\n",
               CgiScriptName, LinkFileName, UriTitle, UriReferer, IconLocation);
         }
      }

      if (Type[0] == 'B')
      {
         /********/
         /* book */
         /********/

         if (!Title[0] || IsSameAsPrevTitle)
            strcpy (HtmlTitle,
                    "<FONT SIZE=-1><I>(Bookreader version)</I></FONT>");

         if (IsOdlShelf || LinkFileName[0] == '/')
         {
            if (IsOdlShelf)
            {
               for (cptr = LinkFileName; *cptr; cptr++);
               while (cptr > LinkFileName && *cptr != '.') cptr--;
               if (*cptr != '.') strcat (LinkFileName, ".bkr");
            }
            else
            {
               for (cptr = LinkFileName; *cptr; cptr++);
               while (cptr > LinkFileName && *cptr != '.') cptr--;
               if (*cptr != '.') strcat (LinkFileName, ".decw$book");
            }

            fprintf (stdout,
"<A HREF=\"%s?title=%s&referer=%s\">\
<IMG SRC=\"%s/book.gif\" ALIGN=top ALT=\"[}{]\"></A>\n",
               LinkFileName,
               UriFormTitle, HyperReaderReferer, IconLocation);
         }
         else
         {
            fprintf (stdout,
"<A HREF=\"%s?book=%s&title=%s&referer=%s\">\
<IMG SRC=\"%s/book.gif\" ALIGN=top ALT=\"[}{]\"></A>\n",
               HyperReaderScriptName, LinkFileName,
               UriFormTitle, HyperReaderReferer, IconLocation);
         }
      }

      if (Type[0] == 'H')
      {
         /****************/
         /* HTML version */
         /****************/

         if (!Title[0] || IsSameAsPrevTitle)
            strcpy (HtmlTitle,
                    "<FONT SIZE=-1><I>(HTML version)</I></FONT>");

         if (IsOdlShelf)
         {
            for (cptr = LinkFileName; *cptr; cptr++);
            while (cptr > LinkFileName && *cptr != '.') cptr--;
            if (*cptr != '.') strcat (LinkFileName, ".htm");
         }
         else
         {
            for (cptr = LinkFileName; *cptr; cptr++);
            while (cptr > LinkFileName && *cptr != '.') cptr--;
            if (*cptr != '.') strcat (LinkFileName, ".html");
         }

         if (IsOdlShelf || LinkFileName[0] == '/')
            cptr = LinkFileName;
         else
            cptr = MapUrl (NULL, LinkFileName, NULL, NULL);

         fprintf (stdout,
"<A HREF=\"%s\">\
<IMG SRC=\"%s/html.gif\" ALIGN=top ALT=\"[HT]\"></A>\n",
           cptr, IconLocation);
      }

      if (Type[0] == 'P' && Type[1] == 'D')
      {
         /*********************************************/
         /* Adobe PDF version (HyperReader extension) */
         /*********************************************/

         if (!Title[0] || IsSameAsPrevTitle)
            strcpy (HtmlTitle,
                    "<FONT SIZE=-1><I>(PDF version)</I></FONT>");

         if (IsOdlShelf || LinkFileName[0] == '/')
            cptr = LinkFileName;
         else
            cptr = MapUrl (NULL, LinkFileName, NULL, NULL);

         fprintf (stdout,
"<A HREF=\"%s\">\
<IMG SRC=\"%s/pdf.gif\" ALIGN=top ALT=\"[PDF]\"></A>\n",
         cptr, IconLocation);
      }

      if (Type[0] == 'P' && Type[1] == 'L')
      {
         /**********************************************/
         /* plain-text version (HyperReader extension) */
         /**********************************************/

         if (!Title[0] || IsSameAsPrevTitle)
            strcpy (HtmlTitle,
                    "<FONT SIZE=-1><I>(plain-text version)</I></FONT>");

         if (IsOdlShelf || LinkFileName[0] == '/')
            cptr = LinkFileName;
         else
            cptr = MapUrl (NULL, LinkFileName, NULL, NULL);

         fprintf (stdout,
"<A HREF=\"%s\">\
<IMG SRC=\"%s/plain.gif\" ALIGN=top ALT=\"[TXT]\"></A>\n",
         cptr, IconLocation);

         if (PrintScriptName[0])
         {
            fprintf (stdout,
"<A HREF=\"%s%s\">\
<IMG SRC=\"%s/print.gif\" ALIGN=top ALT=\"[Prn]\"></A>\n",
            PrintScriptName, cptr, IconLocation);
         }
      }

      if ((Type[0] == 'P' && Type[1] == 'O') ||
          (Type[0] == 'P' && Type[1] == 'S'))
      {
         /**********************************************/
         /* PostScript version (HyperReader extension) */
         /**********************************************/

         if (!Title[0] || IsSameAsPrevTitle)
            strcpy (HtmlTitle,
                    "<FONT SIZE=-1><I>(PostScript version)</I></FONT>");

         if (IsOdlShelf || LinkFileName[0] == '/')
            cptr = LinkFileName;
         else
            cptr = MapUrl (NULL, LinkFileName, NULL, NULL);

         fprintf (stdout,
"<A HREF=\"%s\">\
<IMG SRC=\"%s/ps.gif\" ALIGN=top ALT=\"[PS]\"></A>\n",
         cptr, IconLocation);

         if (PrintScriptName[0])
         {
            fprintf (stdout,
"<A HREF=\"%s%s\">\
<IMG SRC=\"%s/print.gif\" ALIGN=top ALT=\"[Prn]\"></A>\n",
            PrintScriptName, cptr, IconLocation);
         }
      }

      fprintf (stdout, "</NOBR></TD><TD>%s</TD></TR>\n", HtmlTitle);

      strcpy (PrevTitle, Title);
   }

   /****************/
   /* end of shelf */
   /****************/

   sys$close (&ShelfFab, 0, 0);

   if (status == RMS$_EOF) status = SS$_NORMAL;
   if (VMSnok (status))
   {
      CopyIntoHtml (HtmlTitle, FormTitle, -1);
      if (HtmlTitle[0])
         cptr = HtmlTitle;
      else
         cptr = CgiPathInfo;
      ErrorVmsStatus (status, cptr, ShelfSpec, __FILE__, __LINE__);
      return (SS$_NORMAL);
   }

   fprintf (stdout, "</TABLE>\n");
   if (!LineCount) fputs ("<I>Empty bookshelf!</I>\n", stdout);

   fprintf (stdout, "<P><IMG SRC=\"%s/hr.gif\" ALT=\"%s\"><P>\n",
            IconLocation, "+--------------------------------------+");

   if (OriginalReferer[0])
      fprintf (stdout,
"<A HREF=\"%s\"><IMG SRC=\"%s/close.gif\" ALT=\"[Close]\"></A>\n",
               OriginalReferer, IconLocation);

   if (HelpUrl[0])
      fprintf (stdout,
"<A HREF=\"%s\"><IMG SRC=\"%s/help.gif\" ALT=\"[Help]\"></A>\n",
               HelpUrl, IconLocation);

   fprintf (stdout, "</BODY>\n</HTML>\n");
}

/*****************************************************************************/
/*
Copy text from one string to another, converting characters forbidden to 
appear as plain-text in HTML.  For example the '<', '&', etc.  Convert these 
to the corresponding HTML character entities.  Returns number of characters 
copied into HTML string.
*/ 

int CopyIntoHtml
( 
register char *hptr,
register char *sptr,
register int ccnt
)
{
   char  *FirstPtr;

   /*********/
   /* begin */
   /*********/

   FirstPtr = hptr;
   while (*sptr && ccnt--)
   {
      switch (*sptr)
      {
         case '<' : strcpy (hptr, "&lt;"); hptr += 4; sptr++; break;
         case '>' : strcpy (hptr, "&gt;"); hptr += 4; sptr++; break;
         case '&' : strcpy (hptr, "&amp;"); hptr += 5; sptr++; break;
         case '\"' : strcpy (hptr, "&quot;"); hptr += 6; sptr++; break;
         default : if (isprint(*sptr)) *hptr++ = *sptr++; else sptr++;
      }
   }
   *hptr = '\0';
   return (hptr-FirstPtr);
}

/*****************************************************************************/
/*
*/

ErrorGeneral
(
char *Text,
char *SourceFileName,
int SourceLineNumber
)
{
   register char  *cptr;

   /*********/
   /* begin */
   /*********/

   ErrorReported = true;

   /* 
      The source file format provided by the "__FILE__" macro will be
      "device:[directory]name.type;ver".  Reduce that to "name.type".
   */
   for (cptr = SourceFileName; *cptr && *cptr != ';'; cptr++);
   if (*cptr)
   {
      while (*cptr != '.') cptr--;
      *cptr-- = '\0';
   }
   while (*cptr != ']') cptr--;
   cptr++;

   if (HttpHasBeenOutput)
   {
      fprintf (stdout,
"<P><HR SIZE=3>\n\
<H1>ERROR!</H1>\n\
<!-- %s, %s, %d -->\n\
<P>Reported by server.\n\
<P>%s\n\
</BODY>\n\
</HTML>\n",
      SoftwareID, cptr, SourceLineNumber, Text);
   }
   else
   {
      fprintf (stdout,
"%s%s\n\
<HTML>\n\
<HEAD>\n\
<META NAME=\"generator\" CONTENT=\"%s\">\n\
<META NAME=\"module\" CONTENT=\"%s\">\n\
<META NAME=\"line\" CONTENT=\"%d\">\n\
<TITLE>Error 404</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1>ERROR!</H1>\n\
<P>Reported by server.\n\
<P>%s\n\
</BODY>\n\
</HTML>\n",
      Http404Header, SgmlDocType, SoftwareID, cptr, SourceLineNumber, Text);
   }

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
*/

ErrorVmsStatus
(
int StatusValue,
char *Text,
char *HiddenText,
char *SourceFileName,
int SourceLineNumber
)
{
   static char  Message [256];
   static $DESCRIPTOR (MessageDsc, Message);

   register char  *cptr;
   int  status;
   char  *MoreInfoPtr;
   short int  Length;

   /*********/
   /* begin */
   /*********/

   ErrorReported = true;

   if (VMSok (status = sys$getmsg (StatusValue, &Length, &MessageDsc, 1, 0))) 
   {
      Message[Length] = '\0';
      Message[0] = toupper(Message[0]);
   }
   else
      strcpy (Message, "&quot;sys$getmsg() failed&quot;");

   if (StatusValue == RMS$_DEV)
      MoreInfoPtr = "<P>The required CD-ROM may not be mounted.\n";
   else
      MoreInfoPtr = "";

   /* 
      The source file format provided by the "__FILE__" macro will be
      "device:[directory]name.type;ver".  Reduce that to "name.type".
   */
   for (cptr = SourceFileName; *cptr && *cptr != ';'; cptr++);
   if (*cptr)
   {
      while (*cptr != '.') cptr--;
      *cptr-- = '\0';
   }
   while (*cptr != ']') cptr--;
   cptr++;

   if (HttpHasBeenOutput)
   {
      fprintf (stdout,
"<P><HR SIZE=3>\n\
<H1>ERROR!</H1>\n\
<!-- %s, %s, %d -->\n\
<P>Reported by server.\n\
<P>%s ... <TT>%s</TT>\n\
<!-- %%X%08.08X \"%s\" -->\n\
%s\
</BODY>\n\
</HTML>\n",
      SoftwareID, cptr, SourceLineNumber,
      Message, Text, StatusValue, HiddenText, MoreInfoPtr);
   }
   else
   {
      fprintf (stdout,
"%s%s\n\
<HTML>\n\
<HEAD>\n\
<META NAME=\"generator\" CONTENT=\"%s\">\n\
<META NAME=\"module\" CONTENT=\"%s\">\n\
<META NAME=\"line\" CONTENT=\"%d\">\n\
<TITLE>Error 404</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1>ERROR!</H1>\n\
<P>Reported by server.\n\
<P>%s ... <TT>%s</TT>\n\
<!-- %%X%08.08X \"%s\" -->\n\
%s\
</BODY>\n\
</HTML>\n",
      Http404Header, SgmlDocType,
      SoftwareID, cptr, SourceLineNumber,
      Message, Text, StatusValue, HiddenText, MoreInfoPtr);
   }

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
If the object has been modified since the specified date and time then return 
a normal status indicating that the data transfer is to continue.  If not 
modified then send a "not modified" HTTP header and return an error status to 
indicate the object should not be sent.
*/ 
 
int ModifiedSince
(
unsigned long *SinceBinaryTimePtr,
unsigned long *BinaryTimePtr
)
{
   static unsigned long  OneSecondDelta [2] = { -10000000, -1 };
   static char  Http304Header [] = "HTTP/1.0 304 Not modified.\r\n\r\n";

   int  status;
   unsigned long  AdjustedBinTime [2],
                  ScratchBinTime [2];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ModifiedSince()\n");

   /* if request asks for a "reload" (not cached) then give it regardless */
   if (strsame (CgiHttpPragma, "no-cache", -1)) return (SS$_NORMAL);

   /*
      Add one second to the modified time.  Ensures a negative time
      for VMS where fractional seconds may result in inconclusive
      results when the target time is being specified in whole seconds.
   */ 
   if (VMSnok (status =
       lib$add_times (SinceBinaryTimePtr, &OneSecondDelta, &AdjustedBinTime)))
   {
      ErrorVmsStatus (status, "If-Modified-Since:", "", __FILE__, __LINE__);
      return (status);
   }
   if (Debug) fprintf (stdout, "sys$add_times() %%X%08.08X\n", status);

   /* if a positive time results the file has been modified */
   if (VMSok (status =
       lib$sub_times (BinaryTimePtr, &AdjustedBinTime, &ScratchBinTime)))
      return (status);

   if (Debug) fprintf (stdout, "sys$sub_times() %%X%08.08X\n", status);
   if (status != LIB$_NEGTIM)
   {
      ErrorVmsStatus (status, "If-Modified-Since:", "", __FILE__, __LINE__);
      return (status);
   }

   fputs (Http304Header, stdout);

   return (LIB$_NEGTIM);
}

/*****************************************************************************/
/*
Create an HTTP format Greenwich Mean Time (UTC) time string in the storage 
pointed at by 'TimeString', e.g. "Fri, 25 Aug 1995 17:32:40 GMT" (RFC 1123). 
This must be at least 30 characters capacity.  If 'BinTimePtr' is null the 
time string represents the current time.  If it points to a quadword, VMS time 
value the string represents that time.  'TimeString' must point to storage 
large enough for 31 characters.
*/

int HttpGmTimeString
(
char *TimeString,
unsigned long *BinTimePtr
)
{
   static char  *DayNames [] =
      { "", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun" };
   static char  *MonthName [] =
      { "", "Jan", "Feb", "Mar", "Apr", "May", "Jun",
            "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

   static $DESCRIPTOR (HttpTimeFaoDsc, "!AZ, !2ZW !AZ !4ZW !2ZW:!2ZW:!2ZW GMT");
   static $DESCRIPTOR (TimeStringDsc, "");

   int  status;
   unsigned long  BinTime [2],
                  GmTime [2];
   unsigned short  Length;
   unsigned short  NumTime [7];
   unsigned long  DayOfWeek;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HttpGmTimeString()\n");

   if (BinTimePtr == NULL)
      sys$gettim (&GmTime);
   else
   {
      GmTime[0] = BinTimePtr[0];
      GmTime[1] = BinTimePtr[1];
   }
   if (VMSnok (status = TimeAdjustGMT (true, &GmTime)))
      return (status);

   status = sys$numtim (&NumTime, &GmTime);
   if (Debug)
      fprintf (stdout, "sys$numtim() %%X%08.08X %d %d %d %d %d %d %d\n",
               status, NumTime[0], NumTime[1], NumTime[2],
               NumTime[3], NumTime[4], NumTime[5], NumTime[6]);

   if (VMSnok (status = lib$day_of_week (&GmTime, &DayOfWeek)))
      return (status);
   if (Debug)
      fprintf (stdout, "lib$day_of_week() %%X%08.08X is %d\n",
               status, DayOfWeek);

   /* set the descriptor address and size of the resultant time string */
   TimeStringDsc.dsc$w_length = 30;
   TimeStringDsc.dsc$a_pointer = TimeString;

   if (VMSnok (status =
       sys$fao (&HttpTimeFaoDsc, &Length, &TimeStringDsc,
                DayNames[DayOfWeek], NumTime[2], MonthName[NumTime[1]],
                NumTime[0], NumTime[3], NumTime[4], NumTime[5])))
   {
      if (Debug) fprintf (stdout, "sys$fao() %%X%08.08X\n", status);
      TimeString[0] = '\0';
   }
   else
      TimeString[Length] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", TimeString);

   return (status);
}

/*****************************************************************************/
/*
Given a string such as "Fri, 25 Aug 1995 17:32:40 GMT" (RFC 1123), or "Friday,
25-Aug-1995 17:32:40 GMT" (RFC 1036), create an internal, local, binary time
with the current GMT offset.  See complementary function HttpGmTimeString().
*/ 

int HttpGmTime
(
char *TimeString,
unsigned long *BinTimePtr
)
{
   static char  *MonthName [] =
      { "", "Jan", "Feb", "Mar", "Apr", "May", "Jun",
            "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

   register char  *tptr;

   int  status;
   unsigned short  Length;
   unsigned short  NumTime [7] = { 0,0,0,0,0,0,0 };
   unsigned long  DayOfWeek;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "HttpGmTime() |%s|\n", TimeString);

   tptr = TimeString;
   /* hunt straight for the comma after the weekday name! */
   while (*tptr && *tptr != ',') tptr++;
   if (*tptr) tptr++;
   /* span white space between weekday name and date */
   while (*tptr && isspace(*tptr)) tptr++;
   if (Debug) fprintf (stdout, "tptr |%s|\n", tptr);
   if (!*tptr) return (STS$K_ERROR);

   /* get the date and then skip to month name */
   if (isdigit(*tptr)) NumTime[2] = atoi (tptr);
   while (*tptr && isdigit(*tptr)) tptr++;
   while (*tptr && (*tptr == '-' || isspace(*tptr))) tptr++;
   if (Debug) fprintf (stdout, "tptr |%s|\n", tptr);
   if (!*tptr) return (STS$K_ERROR);

   /* get the month number from the name and skip to the year */
   for (NumTime[1] = 1; NumTime[1] <= 12; NumTime[1]++)
      if (strsame (tptr, MonthName[NumTime[1]], 3)) break;
   if (NumTime[1] > 12) return (STS$K_ERROR);
   while (*tptr && isalpha(*tptr)) tptr++;
   while (*tptr && (*tptr == '-' || isspace(*tptr))) tptr++;
   if (Debug) fprintf (stdout, "tptr |%s|\n", tptr);
   if (!*tptr) return (STS$K_ERROR);

   /* get the year and then skip to the hour */
   if (isdigit(*tptr))
   {
      NumTime[0] = atoi (tptr);
      if (NumTime[0] < 100) NumTime[0] += 1900;
   }
   while (*tptr && isdigit(*tptr)) tptr++;
   while (*tptr && isspace(*tptr)) tptr++;
   if (Debug) fprintf (stdout, "tptr |%s|\n", tptr);
   if (!*tptr) return (STS$K_ERROR);

   /* get the hour, minute and second */
   if (isdigit(*tptr)) NumTime[3] = atoi (tptr);
   while (*tptr && isdigit(*tptr)) tptr++;
   if (*tptr == ':') tptr++;
   if (isdigit(*tptr)) NumTime[4] = atoi (tptr);
   while (*tptr && isdigit(*tptr)) tptr++;
   if (*tptr == ':') tptr++;
   if (isdigit(*tptr)) NumTime[5] = atoi (tptr);
   while (*tptr && isdigit(*tptr)) tptr++;
   if (*tptr == ':') tptr++;
   if (!*tptr) return (STS$K_ERROR);

   /* the only thing remaining should be the "GMT" */
   while (*tptr && isspace(*tptr)) tptr++;
   if (Debug) fprintf (stdout, "tptr |%s|\n", tptr);
   if (!strsame (tptr, "GMT", 3)) return (STS$K_ERROR);

   /*******************************************/
   /* convert what looks like legitimate GMT! */
   /*******************************************/

   if (Debug)
      fprintf (stdout, "NumTime[] %d %d %d %d %d %d %d\n",
               NumTime[0], NumTime[1], NumTime[2], NumTime[3], 
               NumTime[4], NumTime[5], NumTime[6]); 
   status = lib$cvt_vectim (&NumTime, BinTimePtr);
   if (VMSnok (status)) return (status);
   if (Debug) fprintf (stdout, "lib$cvt_vectim() %%X%08.08X\n", status);

   return (TimeAdjustGMT (false, BinTimePtr));
}

/*****************************************************************************/
/*
Translate the logical HTTPD$GMT (defined to be something like "+10:30" or "-
01:15") and convert it into a delta time structure and store in 
'TimeGmtDeltaBinary'.  Store whether it is in advance or behind GMT in boolean 
'TimeAheadOfGmt'.  Store the logical string in 'TimeGmtString'.
*/

int TimeSetGMT ()

{
   register char  *cptr, *sptr;

   int  status;
   unsigned short  Length;
   $DESCRIPTOR (GmtLogicalNameDsc, "HTTPD$GMT");
   $DESCRIPTOR (LnmFileDevDsc, "LNM$FILE_DEV");
   $DESCRIPTOR (TimeGmtVmsStringDsc, "");
   struct {
      short int  buf_len;
      short int  item;
      unsigned char   *buf_addr;
      unsigned short  *ret_len;
   } LnmItems [] =
   {
      { sizeof(TimeGmtString)-1, LNM$_STRING, TimeGmtString, &Length },
      { 0,0,0,0 }
   };

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TimeSetGMT()\n");

   status = sys$trnlnm (0, &LnmFileDevDsc, &GmtLogicalNameDsc, 0, &LnmItems);
   if (Debug) fprintf (stdout, "sys$trnlnm() %%X%08.08X\n", status);
   if (VMSnok (status)) return (status);

   TimeGmtString[Length] = '\0';
   if (Debug) fprintf (stdout, "TimeGmtString |%s|\n", TimeGmtString);
   if (*(cptr = TimeGmtString) == '-')
      TimeAheadOfGmt = false;
   else
      TimeAheadOfGmt = true;
   if (*cptr == '+' || *cptr == '-') cptr++;
   sptr = TimeGmtVmsString;
   *sptr++ = '0';
   *sptr++ = ' ';
   while (*cptr) *sptr++ = *cptr++;
   *sptr = '\0';
   if (Debug) fprintf (stdout, "TimeGmtVmsString |%s|\n", TimeGmtVmsString);

   TimeGmtVmsStringDsc.dsc$w_length = sptr - TimeGmtVmsString;
   TimeGmtVmsStringDsc.dsc$a_pointer = TimeGmtVmsString;

   return (sys$bintim (&TimeGmtVmsStringDsc, &TimeGmtDeltaBinary));
}

/*****************************************************************************/
/*
The GMT is generated by calculating using an offset using 
'TimeGmtDeltaOffset' and boolean 'TimeAheadOfGmt'.  Adjust either to or from 
GMT.
*/ 

int TimeAdjustGMT
(
boolean ToGmTime,
unsigned long *BinTimePtr
)
{
   int  status;
   unsigned long  AdjustedTime [2];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "TimeAdjustGMT() ToGmTime: %d\n", ToGmTime);

   if ((ToGmTime && TimeAheadOfGmt) || (!ToGmTime && !TimeAheadOfGmt))
   {
      /* to GMT from local and ahead of GMT, or to local from GMT and behind */
      status = lib$sub_times (BinTimePtr, &TimeGmtDeltaBinary, &AdjustedTime);
      if (Debug) fprintf (stdout, "lib$sub_times() %%X%08.08X\n", status);
   }
   else
   {
      /* to GMT from local and behind GMT, or to local from GMT and ahead */
      status = lib$add_times (BinTimePtr, &TimeGmtDeltaBinary, &AdjustedTime);
      if (Debug) fprintf (stdout, "lib$add_times() %%X%08.08X\n", status);
   }

   if (Debug)
   {
      unsigned short  Length;
      char  String [64];
      $DESCRIPTOR (AdjustedTimeFaoDsc, "AdjustedTime: |!%D|\n");
      $DESCRIPTOR (StringDsc, String);

      sys$fao (&AdjustedTimeFaoDsc, &Length, &StringDsc, &AdjustedTime);
      String[Length] = '\0';
      fputs (String, stdout);
   }

   BinTimePtr[0] = AdjustedTime[0];
   BinTimePtr[1] = AdjustedTime[1];

   return (status);
}

/*****************************************************************************/
/*
Copy text from one string to another, converting characters forbidden to 
appear as plain-text text in an HTTP URL.  For example the '?', '+', '&', etc.  
Convert these to "%nn" hexadecimal escaped characters.  Returns number of
characters copied into URI.
*/ 

int CopyIntoUri
( 
register char *uptr,
register char *sptr,
register int  ccnt
)
{
   char  *FirstPtr;

   /*********/
   /* begin */
   /*********/

   FirstPtr = uptr;
   while (*sptr && ccnt--)
   {  
      switch (*sptr)
      {
         case '\t' : strcpy (uptr, "%09"); uptr += 3; sptr++; break;
         case ' ' : strcpy (uptr, "%20"); uptr += 3; sptr++; break;
         case '!' : strcpy (uptr, "%21"); uptr += 3; sptr++; break;
         case '\"' : strcpy (uptr, "%22"); uptr += 3; sptr++; break;
         case '#' : strcpy (uptr, "%23"); uptr += 3; sptr++; break;
         case '$' : strcpy (uptr, "%24"); uptr += 3; sptr++; break;
         case '%' : strcpy (uptr, "%25"); uptr += 3; sptr++; break;
         case '&' : strcpy (uptr, "%26"); uptr += 3; sptr++; break;
         case '\'' : strcpy (uptr, "%27"); uptr += 3; sptr++; break;
         case '(' : strcpy (uptr, "%28"); uptr += 3; sptr++; break;
         case ')' : strcpy (uptr, "%29"); uptr += 3; sptr++; break;
         case '*' : strcpy (uptr, "%2a"); uptr += 3; sptr++; break;
         case '+' : strcpy (uptr, "%2b"); uptr += 3; sptr++; break;
         case ',' : strcpy (uptr, "%2c"); uptr += 3; sptr++; break;
         case '-' : strcpy (uptr, "%2d"); uptr += 3; sptr++; break;
         case '.' : strcpy (uptr, "%2e"); uptr += 3; sptr++; break;
         case '/' : strcpy (uptr, "%2f"); uptr += 3; sptr++; break;
         case ':' : strcpy (uptr, "%3a"); uptr += 3; sptr++; break;
         case ';' : strcpy (uptr, "%3b"); uptr += 3; sptr++; break;
         case '<' : strcpy (uptr, "%3c"); uptr += 3; sptr++; break;
         case '=' : strcpy (uptr, "%3d"); uptr += 3; sptr++; break;
         case '>' : strcpy (uptr, "%3e"); uptr += 3; sptr++; break;
         case '?' : strcpy (uptr, "%3f"); uptr += 3; sptr++; break;
         case '[' : strcpy (uptr, "%5b"); uptr += 3; sptr++; break;
         case '\\' : strcpy (uptr, "%5c"); uptr += 3; sptr++; break;
         case ']' : strcpy (uptr, "%5d"); uptr += 3; sptr++; break;
         case '^' : strcpy (uptr, "%5e"); uptr += 3; sptr++; break;
         case '_' : strcpy (uptr, "%5f"); uptr += 3; sptr++; break;
         case '{' : strcpy (uptr, "%7b"); uptr += 3; sptr++; break;
         case '|' : strcpy (uptr, "%7c"); uptr += 3; sptr++; break;
         case '}' : strcpy (uptr, "%7d"); uptr += 3; sptr++; break;
         case '~' : strcpy (uptr, "%7e"); uptr += 3; sptr++; break;
         default : if (isprint(*sptr)) *uptr++ = *sptr++; else sptr++;
      }
   }
   *uptr = '\0';
   return (uptr-FirstPtr);
}

/****************************************************************************/
/*
Does a case-insensitive, character-by-character string compare and returns 
true if two strings are the same, or false if not.  If a maximum number of 
characters are specified only those will be compared, if the entire strings 
should be compared then specify the number of characters as 0.
*/ 

boolean strsame
(
register char *sptr1,
register char *sptr2,
register int  count
)
{
   while (*sptr1 && *sptr2)
   {
      if (toupper (*sptr1++) != toupper (*sptr2++)) return (false);
      if (count)
         if (!--count) return (true);
   }
   if (*sptr1 || *sptr2)
      return (false);
   else
      return (true);
}

/*****************************************************************************/
/*
This function provides a simple, uniform mechanism for a C Language script to
access CGI variable values transparently in both the standard CGI and the
CGIplus environments.  Requires the global storage 'CgiPlusEofPtr' to be
non-NULL to indicate a CGIplus environment.

It is completely self-contained. Simply call the function with the CGI
variable name value to be accessed and if that CGI variable exists a pointer
to the value will be returned.  Errors and non-existant variables return NONE
(see below).  The values pointed to should NOT be modified in any way (copy the
value if this is required). 

Supplying an empty variable name (i.e. "") will cause the function to block
until the arrival of a request, upon which it just returns a NONE pointer,
allowing start-of-request synchronization.  This will also release allocated
memory and reset in preparation for reading the next request, an alternative to
calling with a NULL parameter at the end of request processing (see below).

For the CGIplus environment supplying a wilcard variable name (i.e. "*")
returns each CGIplus variable string (line from CGIPLUSIN) in turn, with no
more being indicated by a NONE pointer.

At the end of a CGIplus request call the function with a NULL parameter to
release allocated memory and reset ready for processing the next request's
variables.  It is not necessary to do this if calling with an empty parameter
to synchronize on the start of a request (see above).

The CGIplus functionality works by initially reading all lines from
CGIPLUSIN until the terminating blank line is encountered. These lines are
stored as a series of contiguous null-terminated strings, terminated by an
empty string, in allocated memory. These strings are searched for the required
variable name and a pointer to its value returned if found.

The value NONE shold be set according to the requirements of the application
the function is included within.  If calling functions can handle NULL pointers
being returned then set the value to NULL.  If non-NULL pointers are prefered
then set the value to an empty string.
*/

char* CgiVar (char *VarName)

{
/** #  define NONE NULL **/
#  define NONE ""
#  define CGIPLUS_LINE_SIZE 1024

   static int  VarPtrSize = 0;
   static char  *NextPtr = NULL,
                *VarPtr = NULL;

   register char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "CgiVar() |%s|\n", VarName);

   if (VarName == NULL || !VarName[0])
   {
      /* end of request, release memory and reinitialize */
      if (VarPtr != NULL)
      {
         free (VarPtr);
         NextPtr = VarPtr = NULL;
      }
      if (VarName == NULL) return (NONE);
   }

   if (CgiPlusEofPtr == NULL)
   {
      /* standard CGI environment, get variable from environment */
      if (!VarName[0]) return (NONE);
      if ((cptr = getenv(VarName)) == NULL)
         return (NONE);
      else
         return (cptr);
   }

   if (VarPtr == NULL)
   {
      /* read lines containing CGIplus variables from <CGIPLUSIN> */
      int  LineLength;
      char  Line [CGIPLUS_LINE_SIZE];
      FILE  *CgiPlusIn;
                                                 
      VarPtrSize = 0;
      if ((CgiPlusIn = fopen (getenv("CGIPLUSIN"), "r")) == NULL)
         return (NONE);
      while (fgets (Line, sizeof(Line), CgiPlusIn) != NULL)
      {
         /* first empty line signals the end of CGIplus variables */
         if (Line[0] == '\n') break;
         /* ignore any non-alphabetic line */
         if (!isalpha(Line[0])) continue;
         LineLength = strlen(Line);
         if ((VarPtr = realloc (VarPtr, VarPtrSize+LineLength+2)) == NULL)
            return (NONE);
         memcpy (VarPtr+VarPtrSize, Line, LineLength+1);
         VarPtrSize += LineLength;
         VarPtr[VarPtrSize-1] = '\0';
      }
      fclose (CgiPlusIn);
      if (VarPtr == NULL) return (NONE);
   }

   /* if just waiting for next request return now it's arrived */
   if (!VarName[0]) return (NONE);

   if (VarName[0] == '*')
   {
      /* return each stored line one at a time */
      if (NextPtr == NULL) NextPtr = VarPtr;
      if (!*(sptr = cptr = NextPtr))
      {
         NextPtr = NULL;
         return (NONE);
      }
      while (*cptr) cptr++;
      NextPtr = cptr + 1;
      return (sptr);
   }

   /* search CGIplus variable strings for required variable */
   sptr = VarPtr;
   while (*sptr)
   {
      /* simple comparison between supplied and in-string variable names */
      cptr = VarName;
      while (*cptr && *sptr && *sptr != '=')
      {
         if (toupper(*cptr) != toupper(*sptr)) break;
         cptr++;
         sptr++;
      }
      /* if found then return a pointer to the value */
      if (!*cptr && *sptr == '=') return (sptr+1);
      /* skip to next variable string (if any) */
      while (*sptr) sptr++;
      sptr++;
   }
   /* not found */
   return (NONE);
}

/*****************************************************************************/

